@@ -43,25 +43,20 @@ view_selection_func(GtkTreeSelection *selection, GtkTreeModel *model,
 		     gpointer userdata) {
   tag_context_t *context = (tag_context_t*)userdata;
   GtkTreeIter iter;
-    
+
   if(gtk_tree_model_get_iter(model, &iter, path)) {
     g_assert(gtk_tree_path_get_depth(path) == 1);
 
     tag_t *tag;
     gtk_tree_model_get(model, &iter, TAG_COL_DATA, &tag, -1);
 
-    if(context->but_remove && context->but_edit) {
-      
       /* you just cannot delete or edit the "created_by" tag */
-      if(context->but_remove && context->but_edit) {
-	if(strcasecmp(tag->key, "created_by") == 0) {
-	  gtk_widget_set_sensitive(context->but_remove, FALSE);
-	  gtk_widget_set_sensitive(context->but_edit, FALSE);
-	} else {
-	  gtk_widget_set_sensitive(context->but_remove, TRUE);
-	  gtk_widget_set_sensitive(context->but_edit, TRUE);
-	}
-      }
+    if(strcasecmp(tag->key, "created_by") == 0) {
+      list_button_enable(context->list, LIST_BUTTON_REMOVE, FALSE);
+      list_button_enable(context->list, LIST_BUTTON_EDIT, FALSE);
+    } else {
+      list_button_enable(context->list, LIST_BUTTON_REMOVE, TRUE);
+      list_button_enable(context->list, LIST_BUTTON_EDIT, TRUE);
     }
   }
   
@@ -90,11 +85,10 @@ static void update_collisions(GtkListStore *store, tag_t *tags) {
 }
 
 static void on_tag_remove(GtkWidget *but, tag_context_t *context) {
-  GtkTreeSelection *selection;
   GtkTreeModel     *model;
   GtkTreeIter       iter;
 
-  selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(context->view));
+  GtkTreeSelection *selection = list_get_selection(context->list);
   if(gtk_tree_selection_get_selected(selection, &model, &iter)) {
     tag_t *tag;
     gtk_tree_model_get(model, &iter, TAG_COL_DATA, &tag, -1);
@@ -117,8 +111,8 @@ static void on_tag_remove(GtkWidget *but, tag_context_t *context) {
   }
   
   /* disable remove and edit buttons */
-  gtk_widget_set_sensitive(context->but_remove,  FALSE);
-  gtk_widget_set_sensitive(context->but_edit,  FALSE);
+  list_button_enable(context->list, LIST_BUTTON_REMOVE, FALSE);
+  list_button_enable(context->list, LIST_BUTTON_EDIT, FALSE);
 }
 
 static gboolean tag_edit(tag_context_t *context) {
@@ -127,8 +121,7 @@ static gboolean tag_edit(tag_context_t *context) {
   GtkTreeIter iter;
   tag_t *tag;
 
-  GtkTreeSelection *sel = gtk_tree_view_get_selection(
-		      GTK_TREE_VIEW(context->view));
+  GtkTreeSelection *sel = list_get_selection(context->list);
   if(!sel) {
     printf("got no selection object\n");
     return FALSE;
@@ -218,11 +211,11 @@ static void on_tag_last(GtkWidget *button, tag_context_t *context) {
 	      _("This will overwrite all tags of this %s with the "
 		"ones from the %s selected last.\n\n"
 		"Do you really want this?"),
-	      type_name[context->type], type_name[context->type])) {
+	      type_name[context->object.type], type_name[context->object.type])) {
 
     osm_tags_free(*context->tag);
 
-    if(context->type == NODE)
+    if(context->object.type == NODE)
       *context->tag = osm_tags_copy(context->appdata->map->last_node_tags, TRUE);
     else
       *context->tag = osm_tags_copy(context->appdata->map->last_way_tags, TRUE);
@@ -258,8 +251,8 @@ static void on_tag_add(GtkWidget *button, tag_context_t *context) {
 		     TAG_COL_DATA, *tag,
 		     -1);
 
-  gtk_tree_selection_select_iter(gtk_tree_view_get_selection(
-		     GTK_TREE_VIEW(context->view)), &iter);
+  gtk_tree_selection_select_iter(
+	 list_get_selection(context->list), &iter);
 
   if(!tag_edit(context)) {
     printf("cancelled\n");
@@ -285,46 +278,43 @@ void info_tags_replace(tag_context_t *context) {
 }
 
 static GtkWidget *tag_widget(tag_context_t *context) {
-  GtkWidget *vbox = gtk_vbox_new(FALSE,3);
-  context->view = gtk_tree_view_new();
+  context->list = list_new(LIST_HILDON_WITH_HEADERS_ON_MAEMO5);
 
-  gtk_tree_selection_set_select_function(
-	 gtk_tree_view_get_selection(GTK_TREE_VIEW(context->view)), 
-	 view_selection_func, 
-	 context, NULL);
+  list_set_static_buttons(context->list, G_CALLBACK(on_tag_add), 
+	  G_CALLBACK(on_tag_edit), G_CALLBACK(on_tag_remove), context);
 
-  /* --- "Key" column --- */
-  GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
-  g_object_set(renderer, 
-	       "ellipsize", PANGO_ELLIPSIZE_END, 
-	       "background", "red", 
-	       NULL );
-  GtkTreeViewColumn *column = gtk_tree_view_column_new_with_attributes(
-		 _("Key"), renderer, 
-		 "text", TAG_COL_KEY, 
-		 "background-set", TAG_COL_COLLISION,
-		 NULL);
-  gtk_tree_view_column_set_expand(column, TRUE);
-  gtk_tree_view_insert_column(GTK_TREE_VIEW(context->view), column, -1);
+  list_set_selection_function(context->list, view_selection_func, context);
 
-  /* --- "Value" column --- */
-  renderer = gtk_cell_renderer_text_new();
-  g_object_set(renderer, 
-	       "ellipsize", PANGO_ELLIPSIZE_END, 
-	       NULL );
-  column = gtk_tree_view_column_new_with_attributes(
-		 _("Value"), renderer, 
-		 "text", TAG_COL_VALUE, 
-		 NULL);
-  gtk_tree_view_column_set_expand(column, TRUE);
-  gtk_tree_view_insert_column(GTK_TREE_VIEW(context->view), column, -1);
+  list_set_user_buttons(context->list, 
+			LIST_BUTTON_USER0, _("Last..."), on_tag_last,
+			0);
 
-  /* build and fill the store */
+  /* setup both columns */
+  list_set_columns(context->list, 
+      _("Key"),   TAG_COL_KEY,   
+	   LIST_FLAG_ELLIPSIZE|LIST_FLAG_CAN_HIGHLIGHT, TAG_COL_COLLISION,
+      _("Value"), TAG_COL_VALUE,
+	   LIST_FLAG_ELLIPSIZE,
+      NULL);
+
+  GtkWidget *presets = josm_presets_select(context->appdata, context);
+  if(presets)
+    list_set_custom_user_button(context->list, LIST_BUTTON_USER1, presets);
+
+  /* disable if no appropriate "last" tags have been stored or if the */
+  /* selected item isn't a node or way */
+  if(((context->object.type == NODE) && 
+      (!context->appdata->map->last_node_tags)) ||
+     ((context->object.type == WAY) && 
+      (!context->appdata->map->last_way_tags)) ||
+     ((context->object.type != NODE) && (context->object.type != WAY)))
+	list_button_enable(context->list, LIST_BUTTON_USER0, FALSE);
+
+  /* --------- build and fill the store ------------ */
   context->store = gtk_list_store_new(TAG_NUM_COLS, 
 		G_TYPE_STRING, G_TYPE_STRING, G_TYPE_BOOLEAN, G_TYPE_POINTER);
 
-  gtk_tree_view_set_model(GTK_TREE_VIEW(context->view), 
-			  GTK_TREE_MODEL(context->store));
+  list_set_store(context->list, context->store);
 
   GtkTreeIter iter;
   tag_t *tag = *context->tag;
@@ -342,65 +332,17 @@ static GtkWidget *tag_widget(tag_context_t *context) {
   
   g_object_unref(context->store);
 
-  /* put it into a scrolled window */
-  GtkWidget *scrolled_window = gtk_scrolled_window_new(NULL, NULL);
-  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled_window), 
-				 GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
-  gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(scrolled_window), 
-				      GTK_SHADOW_ETCHED_IN);
-  //  gtk_container_set_border_width(GTK_CONTAINER(scrolled_window), 3);
-  gtk_container_add(GTK_CONTAINER(scrolled_window), context->view);
-  gtk_box_pack_start_defaults(GTK_BOX(vbox), scrolled_window);
+  return context->list;
+}
 
-  /* ------- button box ------------ */
-
-  GtkWidget *hbox = gtk_hbox_new(TRUE,3);
-
-  GtkWidget *but_last = gtk_button_new_with_label(_("Last..."));
-  gtk_box_pack_start_defaults(GTK_BOX(hbox), but_last);
-
-  /* disable if no appropriate "last" tags have been stored or if the */
-  /* selected item isn't a node or way */
-  if(((context->type == NODE) && 
-      (!context->appdata->map->last_node_tags)) ||
-     ((context->type == WAY) && 
-      (!context->appdata->map->last_way_tags)) ||
-     ((context->type != NODE) && (context->type != WAY)))
-    gtk_widget_set_sensitive(but_last, FALSE);
-
-  gtk_signal_connect(GTK_OBJECT(but_last), "clicked",
-  		     GTK_SIGNAL_FUNC(on_tag_last), context);
-
-  GtkWidget *presets = josm_presets_select(context->appdata, context);
-  if(presets) gtk_box_pack_start_defaults(GTK_BOX(hbox), presets);
-
-  context->but_add = gtk_button_new_with_label(_("Add..."));
-  gtk_box_pack_start_defaults(GTK_BOX(hbox), context->but_add);
-  gtk_signal_connect(GTK_OBJECT(context->but_add), "clicked",
-  		     GTK_SIGNAL_FUNC(on_tag_add), context);
-
-  context->but_edit = gtk_button_new_with_label(_("Edit..."));
-  gtk_widget_set_sensitive(context->but_edit, FALSE);
-  gtk_box_pack_start_defaults(GTK_BOX(hbox), context->but_edit);
-  gtk_signal_connect(GTK_OBJECT(context->but_edit), "clicked",
-  		     GTK_SIGNAL_FUNC(on_tag_edit), context);
-
-  context->but_remove = gtk_button_new_with_label(_("Remove"));
-  gtk_widget_set_sensitive(context->but_remove, FALSE);
-  gtk_box_pack_start_defaults(GTK_BOX(hbox), context->but_remove);
-  gtk_signal_connect(GTK_OBJECT(context->but_remove), "clicked",
-		     GTK_SIGNAL_FUNC(on_tag_remove), context);
-
-
-  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
-  return vbox;
+static void on_relation_members(GtkWidget *but, tag_context_t *context) {
+  g_assert(context->object.type == RELATION);
+  relation_show_members(context->dialog, context->object.relation);
 }
 
 /* edit tags of currently selected node or way or of the relation */
 /* given */
-gboolean info_dialog(GtkWidget *parent, appdata_t *appdata, relation_t *relation) {
-  if(!relation)
-    g_assert(appdata->map->selected.type != MAP_TYPE_ILLEGAL);
+gboolean info_dialog(GtkWidget *parent, appdata_t *appdata, object_t *object) {
 
   tag_context_t *context = g_new0(tag_context_t, 1);
   user_t *user = NULL;
@@ -411,41 +353,52 @@ gboolean info_dialog(GtkWidget *parent, appdata_t *appdata, relation_t *relation
   context->appdata = appdata;
   context->tag = &work_copy;
 
-  if(!relation) {
-    switch(appdata->map->selected.type) {
-    case MAP_TYPE_NODE:
-      str = g_strdup_printf(_("Node #%ld"), appdata->map->selected.node->id);
-      user = appdata->map->selected.node->user;
-      work_copy = osm_tags_copy(appdata->map->selected.node->tag, FALSE);
-      stime = appdata->map->selected.node->time;
-      context->type = NODE;
-      context->presets_type = PRESETS_TYPE_NODE;
-      break;
-    case MAP_TYPE_WAY:
-      str = g_strdup_printf(_("Way #%ld"), appdata->map->selected.way->id);
-      user = appdata->map->selected.way->user;
-      work_copy = osm_tags_copy(appdata->map->selected.way->tag, FALSE);
-      stime = appdata->map->selected.way->time;
-      context->type = WAY;
-      context->presets_type = PRESETS_TYPE_WAY;
+  /* use implicit selection if not explicitely given */
+  if(!object) {
+    g_assert((appdata->map->selected.object.type == NODE) ||
+	     (appdata->map->selected.object.type == WAY) ||
+	     (appdata->map->selected.object.type == RELATION));
 
-      if(osm_way_get_last_node(appdata->map->selected.way) == 
-	 osm_way_get_first_node(appdata->map->selected.way))
-	context->presets_type |= PRESETS_TYPE_CLOSEDWAY;
+    context->object = appdata->map->selected.object;
+  } else 
+    context->object = *object;
 
-      break;
-    default:
-      g_assert((appdata->map->selected.type == MAP_TYPE_NODE) ||
-	       (appdata->map->selected.type == MAP_TYPE_WAY));
-      break;
-    }
-  } else {
-    str = g_strdup_printf(_("Relation #%ld"), relation->id);
-    user = relation->user;
-    work_copy = osm_tags_copy(relation->tag, FALSE);
-    stime = relation->time;
-    context->type = RELATION;
+  switch(context->object.type) {
+  case NODE:
+    str = g_strdup_printf(_("Node #" ITEM_ID_FORMAT), context->object.node->id);
+    user = context->object.node->user;
+    work_copy = osm_tags_copy(context->object.node->tag, FALSE);
+    stime = context->object.node->time;
+    context->presets_type = PRESETS_TYPE_NODE;
+    break;
+
+  case WAY:
+    str = g_strdup_printf(_("Way #" ITEM_ID_FORMAT), context->object.way->id);
+    user = context->object.way->user;
+    work_copy = osm_tags_copy(context->object.way->tag, FALSE);
+    stime = context->object.way->time;
+    context->presets_type = PRESETS_TYPE_WAY;
+    
+    if(osm_way_get_last_node(context->object.way) == 
+       osm_way_get_first_node(context->object.way))
+      context->presets_type |= PRESETS_TYPE_CLOSEDWAY;
+    
+    break;
+    
+  case RELATION:
+    str = g_strdup_printf(_("Relation #" ITEM_ID_FORMAT), 
+			  context->object.relation->id);
+    user = context->object.relation->user;
+    work_copy = osm_tags_copy(context->object.relation->tag, FALSE);
+    stime = context->object.relation->time;
     context->presets_type = PRESETS_TYPE_RELATION;
+    break;
+
+  default:
+    g_assert((context->object.type == NODE) ||
+	     (context->object.type == WAY) ||
+	     (context->object.type == RELATION));
+    break;
   }
 
   context->dialog = gtk_dialog_new_with_buttons(str,
@@ -460,66 +413,67 @@ gboolean info_dialog(GtkWidget *parent, appdata_t *appdata, relation_t *relation
 
   /* making the dialog a little wider makes it less "crowded" */
 #ifdef USE_HILDON
-  gtk_window_set_default_size(GTK_WINDOW(context->dialog), 500, 300);
+  gtk_window_set_default_size(GTK_WINDOW(context->dialog), 440, 400);
 #else
-  gtk_window_set_default_size(GTK_WINDOW(context->dialog), 400, 200);
+  // Conversely, desktop builds should display a little narrower
+  gtk_window_set_default_size(GTK_WINDOW(context->dialog), 400, 300);
 #endif
 
   GtkWidget *label;
   GtkWidget *table = gtk_table_new(2, 2, FALSE);  // x, y
 
   /* ------------ user ----------------- */
-  char *u_str = NULL;
-  if(user) u_str = g_strdup_printf(_("User: %s"), user->name);
-  else     u_str = g_strdup_printf(_("User: ---"));
-  label = gtk_label_new(u_str);
-  gtk_table_attach_defaults(GTK_TABLE(table),  label, 0, 1, 0, 1);
-  g_free(u_str);
+  if(user) {
+    label = gtk_label_new(user->name);
+    gtk_label_set_ellipsize(GTK_LABEL(label), PANGO_ELLIPSIZE_END);
+    gtk_table_attach_defaults(GTK_TABLE(table),  label, 0, 1, 0, 1);
+  }
 
   /* ------------ time ----------------- */
 
   struct tm *loctime = localtime(&stime);
   char time_str[32];
   strftime(time_str, sizeof(time_str), "%x %X", loctime);
-  char *t_str = g_strdup_printf(_("Time: %s"), time_str);
-  label = gtk_label_new(t_str);
-  g_free(t_str);
+  label = gtk_label_new(time_str);
   gtk_table_attach_defaults(GTK_TABLE(table),  label, 1, 2, 0, 1);
 
   /* ------------ coordinate (only for nodes) ----------------- */
-  if(!relation) {
-    if(appdata->map->selected.type == MAP_TYPE_NODE) {
-      char pos_str[32];
-      pos_lat_str(pos_str, sizeof(pos_str),appdata->map->selected.node->pos.lat);
-      label = gtk_label_new(pos_str);
-      gtk_table_attach_defaults(GTK_TABLE(table),  label, 0, 1, 1, 2);
-      pos_lat_str(pos_str, sizeof(pos_str),appdata->map->selected.node->pos.lon);
-      label = gtk_label_new(pos_str);
-      gtk_table_attach_defaults(GTK_TABLE(table),  label, 1, 2, 1, 2);
-    } else {
-      char *nodes_str = g_strdup_printf(_("Length: %u nodes"), 
-		osm_way_number_of_nodes(appdata->map->selected.way));
-      label = gtk_label_new(nodes_str);
-      gtk_table_attach_defaults(GTK_TABLE(table),  label, 0, 1, 1, 2);
-      g_free(nodes_str);
+  switch(context->object.type) {
+  case NODE: {
+    char pos_str[32];
+    pos_lat_str(pos_str, sizeof(pos_str), context->object.node->pos.lat);
+    label = gtk_label_new(pos_str);
+    gtk_table_attach_defaults(GTK_TABLE(table),  label, 0, 1, 1, 2);
+    pos_lat_str(pos_str, sizeof(pos_str), context->object.node->pos.lon);
+    label = gtk_label_new(pos_str);
+    gtk_table_attach_defaults(GTK_TABLE(table),  label, 1, 2, 1, 2);
+  } break;
 
-      char *type_str = g_strdup_printf("%s (%s)",
-	 (osm_way_get_last_node(appdata->map->selected.way) == 
-	  osm_way_get_first_node(appdata->map->selected.way))?
-				       "closed way":"open way",
-	 (appdata->map->selected.way->draw.flags & OSM_DRAW_FLAG_AREA)?
-				       "area":"line");
+  case WAY: {
+    char *nodes_str = g_strdup_printf(_("Length: %u nodes"), 
+	      osm_way_number_of_nodes(context->object.way));
+    label = gtk_label_new(nodes_str);
+    gtk_table_attach_defaults(GTK_TABLE(table),  label, 0, 1, 1, 2);
+    g_free(nodes_str);
+    
+    char *type_str = g_strdup_printf("%s (%s)",
+     (osm_way_get_last_node(context->object.way) == 
+      osm_way_get_first_node(context->object.way))?
+			     "closed way":"open way",
+	     (context->object.way->draw.flags & OSM_DRAW_FLAG_AREA)?
+			       "area":"line");
  
-      label = gtk_label_new(type_str);      
-      gtk_table_attach_defaults(GTK_TABLE(table),  label, 1, 2, 1, 2);
-      g_free(type_str);
-    }
-  } else {
+    label = gtk_label_new(type_str);      
+    gtk_table_attach_defaults(GTK_TABLE(table),  label, 1, 2, 1, 2);
+    g_free(type_str);
+  } break;
+
+  case RELATION: {
     /* relations tell something about their members */
     gint nodes = 0, ways = 0, relations = 0;
-    member_t *member = relation->member;
+    member_t *member = context->object.relation->member;
     while(member) {
-      switch(member->type) {
+      switch(member->object.type) {
       case NODE:
       case NODE_ID:
 	nodes++;
@@ -532,7 +486,7 @@ gboolean info_dialog(GtkWidget *parent, appdata_t *appdata, relation_t *relation
       case RELATION_ID:
 	relations++;
 	break;
-
+	
       default:
 	break;
       }
@@ -540,12 +494,23 @@ gboolean info_dialog(GtkWidget *parent, appdata_t *appdata, relation_t *relation
       member = member->next;
     }
 
-    char *str = g_strdup_printf(_("Members: %d nodes, %d ways, %d relations"),
-				nodes, ways, relations);
+    char *str = 
+      g_strdup_printf(_("Members: %d nodes, %d ways, %d relations"),
+		      nodes, ways, relations);
+
+    GtkWidget *member_btn = gtk_button_new_with_label(str);
+    gtk_signal_connect(GTK_OBJECT(member_btn), "clicked", 
+		       GTK_SIGNAL_FUNC(on_relation_members), context);
+    gtk_table_attach_defaults(GTK_TABLE(table), member_btn, 0, 2, 1, 2);
 
-    gtk_table_attach_defaults(GTK_TABLE(table), gtk_label_new(str), 0, 2, 1, 2);
     g_free(str);
-  }
+    break;
+
+  default:
+    printf("ERROR: No node, way or relation\n");
+    g_assert(0);
+    break;
+  } }
 
   gtk_box_pack_start(GTK_BOX(GTK_DIALOG(context->dialog)->vbox), table, 
 		     FALSE, FALSE, 0);
@@ -566,38 +531,39 @@ gboolean info_dialog(GtkWidget *parent, appdata_t *appdata, relation_t *relation
 
     gtk_widget_destroy(context->dialog);
 
-    if(!relation) {
-      /* replace original tags with work copy */
-      switch(appdata->map->selected.type) {
+    /* replace original tags with work copy */
+    switch(context->object.type) {
 
-      case MAP_TYPE_NODE:
-	osm_tags_free(appdata->map->selected.node->tag);
-	appdata->map->selected.node->tag = osm_tags_copy(work_copy, TRUE);
-	break;
+    case NODE:
+      osm_tags_free(context->object.node->tag);
+      context->object.node->tag = osm_tags_copy(work_copy, TRUE);
+      break;
 	
-      case MAP_TYPE_WAY:
-	osm_tags_free(appdata->map->selected.way->tag);
-	appdata->map->selected.way->tag = osm_tags_copy(work_copy, TRUE);
-	break;
+    case WAY:
+      osm_tags_free(context->object.way->tag);
+      context->object.way->tag = osm_tags_copy(work_copy, TRUE);
+      break;
 	
-      default:
-	break;
-      }
-
-      /* since nodes being parts of ways but with no tags are invisible, */
-      /* the result of editing them may have changed their visibility */
-      map_item_redraw(appdata, &appdata->map->selected);
-      map_item_set_flags(&context->appdata->map->selected, OSM_FLAG_DIRTY, 0);
-    } else {
-      osm_tags_free(relation->tag);
-      relation->tag = osm_tags_copy(work_copy, TRUE);
-      relation->flags |= OSM_FLAG_DIRTY;
+    case RELATION:
+      osm_tags_free(context->object.relation->tag);
+      context->object.relation->tag = osm_tags_copy(work_copy, TRUE);
+      break;
+      
+    default:
+      break;
     }
+    
+    /* since nodes being parts of ways but with no tags are invisible, */
+    /* the result of editing them may have changed their visibility */
+    if(!object && context->object.type != RELATION) 
+      map_item_redraw(appdata, &appdata->map->selected);
+
+    osm_object_set_flags(&context->object, OSM_FLAG_DIRTY, 0);
   } else {
     gtk_widget_destroy(context->dialog);
     osm_tags_free(work_copy);
   }
-
+  
   g_free(context);
   return ok;
 }
