@@ -35,12 +35,6 @@
 #error "Tree not enabled in libxml"
 #endif
 
-/* determine where a node/way/relation read from the osm file */
-/* is inserted into the internal database */
-// #define OSM_SORT_ID
-#define OSM_SORT_LAST
-// #define OSM_SORT_FIRST
-
 /* ------------------------- bounds handling --------------------- */
 
 static void osm_bounds_free(bounds_t *bounds) {
@@ -233,7 +227,7 @@ void osm_node_dump(node_t *node) {
   char buf[64];
   struct tm tm;
     
-  printf("Id:      %lu\n", node->id);
+  printf("Id:      "ITEM_ID_FORMAT"\n", node->id);
   printf("User:    %s\n", node->user?node->user->name:"<unspecified>");
   printf("Visible: %s\n", node->visible?"yes":"no");
   
@@ -266,7 +260,7 @@ void osm_node_chain_free(node_chain_t *node_chain) {
 }
 
 void osm_way_free(way_t *way) {
-  //  printf("freeing way #%ld\n", way->id);
+  //  printf("freeing way #" ITEM_ID_FORMAT "\n", way->id);
 
   osm_node_chain_free(way->node_chain);
   osm_tags_free(way->tag);
@@ -311,12 +305,12 @@ void osm_way_dump(way_t *way) {
   char buf[64];
   struct tm tm;
 
-  printf("Id:      %lu\n", way->id);
+  printf("Id:      "ITEM_ID_FORMAT"\n", way->id);
   printf("User:    %s\n", way->user?way->user->name:"<unspecified>");
   printf("Visible: %s\n", way->visible?"yes":"no");
   node_chain_t *node_chain = way->node_chain;
   while(node_chain) {
-    printf("  Node:  %lu\n", node_chain->node->id);
+    printf("  Node:  "ITEM_ID_FORMAT"\n", node_chain->node->id);
     node_chain = node_chain->next;
   }
   
@@ -345,7 +339,7 @@ node_chain_t *osm_parse_osm_way_nd(osm_t *osm,
 
     /* search matching node */
     node_chain->node = osm_get_node_by_id(osm, id);
-    if(!node_chain->node) printf("Node id %lu not found\n", id);
+    if(!node_chain->node) printf("Node id " ITEM_ID_FORMAT " not found\n", id);
     else                  node_chain->node->ways++;
 
     xmlFree(prop);
@@ -392,14 +386,14 @@ void osm_relations_dump(relation_t *relation) {
     char buf[64];
     struct tm tm;
 
-    printf("Id:      %lu\n", relation->id);
+    printf("Id:      "ITEM_ID_FORMAT"\n", relation->id);
     printf("User:    %s\n", 
 	   relation->user?relation->user->name:"<unspecified>");
     printf("Visible: %s\n", relation->visible?"yes":"no");
 
     member_t *member = relation->member;
     while(member) {
-      switch(member->type) {
+      switch(member->object.type) {
       case ILLEGAL:
       case NODE_ID:
       case WAY_ID:
@@ -407,21 +401,21 @@ void osm_relations_dump(relation_t *relation) {
 	break;
 
       case NODE:
-	if(member->node)
-	  printf(" Member: Node, id = %lu, role = %s\n", 
-		 member->node->id, member->role);
+	if(member->object.node)
+	  printf(" Member: Node, id = " ITEM_ID_FORMAT ", role = %s\n", 
+		 member->object.node->id, member->role);
 	break;
 
       case WAY:
-	if(member->way)
-	printf(" Member: Way, id = %lu, role = %s\n", 
-	       member->way->id, member->role);
+	if(member->object.way)
+	  printf(" Member: Way, id = " ITEM_ID_FORMAT ", role = %s\n", 
+		 member->object.way->id, member->role);
 	break;
-
+	
       case RELATION:
-	if(member->relation)
-	printf(" Member: Relation, id = %lu, role = %s\n", 
-	       member->relation->id, member->role);
+	if(member->object.relation)
+	  printf(" Member: Relation, id = " ITEM_ID_FORMAT ", role = %s\n", 
+		 member->object.relation->id, member->role);
 	break;
       }
 
@@ -442,47 +436,47 @@ member_t *osm_parse_osm_relation_member(osm_t *osm,
 			  xmlDocPtr doc, xmlNode *a_node) {
   char *prop;
   member_t *member = g_new0(member_t, 1);
-  member->type = ILLEGAL;
+  member->object.type = ILLEGAL;
 
   if((prop = (char*)xmlGetProp(a_node, (unsigned char*)"type"))) {
-    if(strcasecmp(prop, "way") == 0)           member->type = WAY;
-    else if(strcasecmp(prop, "node") == 0)     member->type = NODE;
-    else if(strcasecmp(prop, "relation") == 0) member->type = RELATION;
+    if(strcasecmp(prop, "way") == 0)           member->object.type = WAY;
+    else if(strcasecmp(prop, "node") == 0)     member->object.type = NODE;
+    else if(strcasecmp(prop, "relation") == 0) member->object.type = RELATION;
     xmlFree(prop);
   }
 
   if((prop = (char*)xmlGetProp(a_node, (unsigned char*)"ref"))) {
     item_id_t id = strtoul(prop, NULL, 10);
 
-    switch(member->type) {
+    switch(member->object.type) {
     case ILLEGAL:
       printf("Unable to store illegal type\n");
       break;
 
     case WAY:
       /* search matching way */
-      member->way = osm_get_way_by_id(osm, id);
-      if(!member->way) {
-	member->type = WAY_ID;
-	member->id = id;
+      member->object.way = osm_get_way_by_id(osm, id);
+      if(!member->object.way) {
+	member->object.type = WAY_ID;
+	member->object.id = id;
       }
       break;
 
     case NODE:
       /* search matching node */
-      member->node = osm_get_node_by_id(osm, id);
-      if(!member->node) {
-	member->type = NODE_ID;
-	member->id = id;
+      member->object.node = osm_get_node_by_id(osm, id);
+      if(!member->object.node) {
+	member->object.type = NODE_ID;
+	member->object.id = id;
       }
       break;
 
     case RELATION:
       /* search matching relation */
-      member->relation = osm_get_relation_by_id(osm, id);
-      if(!member->relation) {
-	member->type = NODE_ID;
-	member->id = id;
+      member->object.relation = osm_get_relation_by_id(osm, id);
+      if(!member->object.relation) {
+	member->object.type = NODE_ID;
+	member->object.id = id;
       }
       break;
 
@@ -685,6 +679,12 @@ static node_t *process_node(xmlTextReaderPtr reader, osm_t *osm) {
     xmlFree(prop);
   }
 
+  /* new in api 0.6: */
+  if((prop = (char*)xmlTextReaderGetAttribute(reader, BAD_CAST "version"))) {
+    node->version = strtoul(prop, NULL, 10);
+    xmlFree(prop);
+  }
+
   if((prop = (char*)xmlTextReaderGetAttribute(reader, BAD_CAST "lat"))) {
     node->pos.lat = g_ascii_strtod(prop, NULL);
     xmlFree(prop);
@@ -759,7 +759,7 @@ static node_chain_t *process_nd(xmlTextReaderPtr reader, osm_t *osm) {
     
     /* search matching node */
     node_chain->node = osm_get_node_by_id(osm, id);
-    if(!node_chain->node) printf("Node id %lu not found\n", id);
+    if(!node_chain->node) printf("Node id " ITEM_ID_FORMAT " not found\n", id);
     else                  node_chain->node->ways++;
 
     xmlFree(prop);
@@ -782,6 +782,12 @@ static way_t *process_way(xmlTextReaderPtr reader, osm_t *osm) {
     xmlFree(prop);
   }
 
+  /* new in api 0.6: */
+  if((prop = (char*)xmlTextReaderGetAttribute(reader, BAD_CAST "version"))) {
+    way->version = strtoul(prop, NULL, 10);
+    xmlFree(prop);
+  }
+
   if((prop = (char*)xmlTextReaderGetAttribute(reader, BAD_CAST "user"))) {
     way->user = osm_user(osm, prop);
     xmlFree(prop);
@@ -842,47 +848,47 @@ static way_t *process_way(xmlTextReaderPtr reader, osm_t *osm) {
 static member_t *process_member(xmlTextReaderPtr reader, osm_t *osm) {
   char *prop;
   member_t *member = g_new0(member_t, 1);
-  member->type = ILLEGAL;
+  member->object.type = ILLEGAL;
 
   if((prop = (char*)xmlTextReaderGetAttribute(reader, BAD_CAST "type"))) {
-    if(strcasecmp(prop, "way") == 0)           member->type = WAY;
-    else if(strcasecmp(prop, "node") == 0)     member->type = NODE;
-    else if(strcasecmp(prop, "relation") == 0) member->type = RELATION;
+    if(strcasecmp(prop, "way") == 0)           member->object.type = WAY;
+    else if(strcasecmp(prop, "node") == 0)     member->object.type = NODE;
+    else if(strcasecmp(prop, "relation") == 0) member->object.type = RELATION;
     xmlFree(prop);
   }
 
   if((prop = (char*)xmlTextReaderGetAttribute(reader, BAD_CAST "ref"))) {
     item_id_t id = strtoul(prop, NULL, 10);
 
-    switch(member->type) {
+    switch(member->object.type) {
     case ILLEGAL:
       printf("Unable to store illegal type\n");
       break;
 
     case WAY:
       /* search matching way */
-      member->way = osm_get_way_by_id(osm, id);
-      if(!member->way) {
-	member->type = WAY_ID;
-	member->id = id;
+      member->object.way = osm_get_way_by_id(osm, id);
+      if(!member->object.way) {
+	member->object.type = WAY_ID;
+	member->object.id = id;
       }
       break;
 
     case NODE:
       /* search matching node */
-      member->node = osm_get_node_by_id(osm, id);
-      if(!member->node) {
-	member->type = NODE_ID;
-	member->id = id;
+      member->object.node = osm_get_node_by_id(osm, id);
+      if(!member->object.node) {
+	member->object.type = NODE_ID;
+	member->object.id = id;
       }
       break;
 
     case RELATION:
       /* search matching relation */
-      member->relation = osm_get_relation_by_id(osm, id);
-      if(!member->relation) {
-	member->type = NODE_ID;
-	member->id = id;
+      member->object.relation = osm_get_relation_by_id(osm, id);
+      if(!member->object.relation) {
+	member->object.type = NODE_ID;
+	member->object.id = id;
       }
       break;
 
@@ -913,6 +919,12 @@ static relation_t *process_relation(xmlTextReaderPtr reader, osm_t *osm) {
     xmlFree(prop);
   }
 
+  /* new in api 0.6: */
+  if((prop = (char*)xmlTextReaderGetAttribute(reader, BAD_CAST "version"))) {
+    relation->version = strtoul(prop, NULL, 10);
+    xmlFree(prop);
+  }
+
   if((prop = (char*)xmlTextReaderGetAttribute(reader, BAD_CAST "user"))) {
     relation->user = osm_user(osm, prop);
     xmlFree(prop);
@@ -1156,6 +1168,7 @@ gboolean osm_node_has_value(node_t *node, char *str) {
 gboolean osm_node_has_tag(node_t *node) {
   tag_t *tag = node->tag;
 
+  /* created_by tags don't count as real tags */
   if(tag && strcasecmp(tag->key, "created_by") == 0)
     tag = tag->next;
 
@@ -1178,8 +1191,10 @@ static void osm_generate_tags(tag_t *tag, xmlNodePtr node) {
   while(tag) {
     /* make sure "created_by" tag contains our id */
     if(strcasecmp(tag->key, "created_by") == 0) {
-      g_free(tag->value);
-      tag->value = g_strdup(PACKAGE " v" VERSION);
+      if(strcasecmp(tag->value, PACKAGE " v" VERSION) != 0) {      
+	g_free(tag->value);
+	tag->value = g_strdup(PACKAGE " v" VERSION);
+      }
     }
 
     xmlNodePtr tag_node = xmlNewChild(node, NULL, BAD_CAST "tag", NULL);
@@ -1190,7 +1205,8 @@ static void osm_generate_tags(tag_t *tag, xmlNodePtr node) {
 }
 
 /* build xml representation for a way */
-char *osm_generate_xml(osm_t *osm, type_t type, void *item) {
+static char *osm_generate_xml(osm_t *osm, item_id_t changeset, 
+		       type_t type, void *item) {
   char str[32];
   xmlChar *result = NULL;
   int len = 0;
@@ -1199,8 +1215,10 @@ char *osm_generate_xml(osm_t *osm, type_t type, void *item) {
 
   xmlDocPtr doc = xmlNewDoc(BAD_CAST "1.0");
   xmlNodePtr root_node = xmlNewNode(NULL, BAD_CAST "osm");
+#ifndef API06
   xmlNewProp(root_node, BAD_CAST "version", BAD_CAST "0.5");
-  xmlNewProp(root_node, BAD_CAST "generator", BAD_CAST PACKAGE " V" VERSION);
+  xmlNewProp(root_node, BAD_CAST "generator", BAD_CAST PACKAGE " v" VERSION);
+#endif
   xmlDocSetRootElement(doc, root_node);
 
   switch(type) {
@@ -1214,9 +1232,15 @@ char *osm_generate_xml(osm_t *osm, type_t type, void *item) {
 	snprintf(str, sizeof(str), "%u", (unsigned)node->id);
 	xmlNewProp(node_node, BAD_CAST "id", BAD_CAST str);
       }
-      g_ascii_dtostr(str, sizeof(str), node->pos.lat);
+#ifdef API06
+      snprintf(str, sizeof(str), "%u", (unsigned)node->version);
+      xmlNewProp(node_node, BAD_CAST "version", BAD_CAST str);
+      snprintf(str, sizeof(str), "%u", (unsigned)changeset);
+      xmlNewProp(node_node, BAD_CAST "changeset", BAD_CAST str);
+#endif
+      g_ascii_formatd(str, sizeof(str), LL_FORMAT, node->pos.lat);
       xmlNewProp(node_node, BAD_CAST "lat", BAD_CAST str);
-      g_ascii_dtostr(str, sizeof(str), node->pos.lon);
+      g_ascii_formatd(str, sizeof(str), LL_FORMAT, node->pos.lon);
       xmlNewProp(node_node, BAD_CAST "lon", BAD_CAST str);    
       osm_generate_tags(node->tag, node_node);
     }
@@ -1228,11 +1252,17 @@ char *osm_generate_xml(osm_t *osm, type_t type, void *item) {
       xmlNodePtr way_node = xmlNewChild(root_node, NULL, BAD_CAST "way", NULL);
       snprintf(str, sizeof(str), "%u", (unsigned)way->id);
       xmlNewProp(way_node, BAD_CAST "id", BAD_CAST str);
+#ifdef API06
+      snprintf(str, sizeof(str), "%u", (unsigned)way->version);
+      xmlNewProp(way_node, BAD_CAST "version", BAD_CAST str);
+      snprintf(str, sizeof(str), "%u", (unsigned)changeset);
+      xmlNewProp(way_node, BAD_CAST "changeset", BAD_CAST str);
+#endif
       
       node_chain_t *node_chain = way->node_chain;
       while(node_chain) {
 	xmlNodePtr nd_node = xmlNewChild(way_node, NULL, BAD_CAST "nd", NULL);
-	char *str = g_strdup_printf("%ld", node_chain->node->id);
+	char *str = g_strdup_printf(ITEM_ID_FORMAT, node_chain->node->id);
 	xmlNewProp(nd_node, BAD_CAST "ref", BAD_CAST str);
 	g_free(str);
 	node_chain = node_chain->next;
@@ -1249,26 +1279,32 @@ char *osm_generate_xml(osm_t *osm, type_t type, void *item) {
 					BAD_CAST "relation", NULL);
       snprintf(str, sizeof(str), "%u", (unsigned)relation->id);
       xmlNewProp(rel_node, BAD_CAST "id", BAD_CAST str);
+#ifdef API06
+      snprintf(str, sizeof(str), "%u", (unsigned)relation->version);
+      xmlNewProp(rel_node, BAD_CAST "version", BAD_CAST str);
+      snprintf(str, sizeof(str), "%u", (unsigned)changeset);
+      xmlNewProp(rel_node, BAD_CAST "changeset", BAD_CAST str);
+#endif
       
       member_t *member = relation->member;
       while(member) {
 	xmlNodePtr m_node = xmlNewChild(rel_node,NULL,BAD_CAST "member", NULL);
 	char *str = NULL;
 
-	switch(member->type) {
+	switch(member->object.type) {
 	case NODE:
 	  xmlNewProp(m_node, BAD_CAST "type", BAD_CAST "node");
-	  str = g_strdup_printf("%ld", member->node->id);
+	  str = g_strdup_printf(ITEM_ID_FORMAT, member->object.node->id);
 	  break;
 
 	case WAY:
 	  xmlNewProp(m_node, BAD_CAST "type", BAD_CAST "way");
-	  str = g_strdup_printf("%ld", member->way->id);
+	  str = g_strdup_printf(ITEM_ID_FORMAT, member->object.way->id);
 	  break;
 
 	case RELATION:
 	  xmlNewProp(m_node, BAD_CAST "type", BAD_CAST "relation");
-	  str = g_strdup_printf("%ld", member->relation->id);
+	  str = g_strdup_printf(ITEM_ID_FORMAT, member->object.relation->id);
 	  break;
 
 	default:
@@ -1308,20 +1344,52 @@ char *osm_generate_xml(osm_t *osm, type_t type, void *item) {
 }
 
 /* build xml representation for a node */
-char *osm_generate_xml_node(osm_t *osm, node_t *node) {
-  return osm_generate_xml(osm, NODE, node);
+char *osm_generate_xml_node(osm_t *osm, item_id_t changeset, node_t *node) {
+  return osm_generate_xml(osm, changeset, NODE, node);
 }
 
 /* build xml representation for a way */
-char *osm_generate_xml_way(osm_t *osm, way_t *way) {
-  return osm_generate_xml(osm, WAY, way);
+char *osm_generate_xml_way(osm_t *osm, item_id_t changeset, way_t *way) {
+  return osm_generate_xml(osm, changeset, WAY, way);
 }
 
 /* build xml representation for a relation */
-char *osm_generate_xml_relation(osm_t *osm, relation_t *relation) {
-  return osm_generate_xml(osm, RELATION, relation);
+char *osm_generate_xml_relation(osm_t *osm, item_id_t changeset, 
+				relation_t *relation) {
+  return osm_generate_xml(osm, changeset, RELATION, relation);
 }
 
+/* build xml representation for a changeset */
+char *osm_generate_xml_changeset(osm_t *osm, char *comment) {
+  xmlChar *result = NULL;
+  int len = 0;
+
+  /* tags for this changeset */
+  tag_t tag_comment = { 
+    .key = "comment", .value = comment, .next = NULL };
+  tag_t tag_creator = { 
+    .key = "created_by", .value = PACKAGE " v" VERSION, .next = &tag_comment };
+
+  LIBXML_TEST_VERSION;
+
+  xmlDocPtr doc = xmlNewDoc(BAD_CAST "1.0");
+  xmlNodePtr root_node = xmlNewNode(NULL, BAD_CAST "osm");
+  xmlDocSetRootElement(doc, root_node);
+
+  xmlNodePtr cs_node = xmlNewChild(root_node, NULL, BAD_CAST "changeset", NULL);
+  osm_generate_tags(&tag_creator, cs_node);
+
+  xmlDocDumpFormatMemoryEnc(doc, &result, &len, "UTF-8", 1);
+  xmlFreeDoc(doc);
+  xmlCleanupParser();
+
+  //  puts("xml encoding result:");
+  //  puts((char*)result);
+
+  return (char*)result;
+}
+
+
 /* the following three functions are eating much CPU power */
 /* as they search the objects lists. Hashing is supposed to help */
 node_t *osm_get_node_by_id(osm_t *osm, item_id_t id) {
@@ -1457,15 +1525,18 @@ node_t *osm_node_new(osm_t *osm, gint x, gint y) {
   printf("Creating new node\n");
 
   node_t *node = g_new0(node_t, 1);
+  node->version = 1;
   node->lpos.x = x; 
   node->lpos.y = y;
   node->visible = TRUE;
   node->time = time(NULL);
 
+#ifndef API06
   /* add created_by tag */
   node->tag = g_new0(tag_t, 1);
   node->tag->key = g_strdup("created_by");
   node->tag->value = g_strdup(PACKAGE " v" VERSION);
+#endif
 
   /* convert screen position back to ll */
   lpos2pos(osm->bounds, &node->lpos, &node->pos);
@@ -1502,14 +1573,17 @@ way_t *osm_way_new(void) {
   printf("Creating new way\n");
 
   way_t *way = g_new0(way_t, 1);
+  way->version = 1;
   way->visible = TRUE;
   way->flags = OSM_FLAG_NEW;
   way->time = time(NULL);
 
+#ifndef API06
   /* add created_by tag */
   way->tag = g_new0(tag_t, 1);
   way->tag->key = g_strdup("created_by");
   way->tag->value = g_strdup(PACKAGE " v" VERSION);
+#endif
 
   return way;
 }
@@ -1534,8 +1608,8 @@ way_chain_t *osm_node_delete(osm_t *osm, icon_t **icon,
 
   /* new nodes aren't stored on the server and are just deleted permanently */
   if(node->flags & OSM_FLAG_NEW) {
-    printf("About to delete NEW node #%ld -> force permanent delete\n",
-	   node->id);
+    printf("About to delete NEW node #" ITEM_ID_FORMAT 
+	   " -> force permanent delete\n", node->id);
     permanently = TRUE;
   }
 
@@ -1571,10 +1645,10 @@ way_chain_t *osm_node_delete(osm_t *osm, icon_t **icon,
   }
 
   if(!permanently) {
-    printf("mark node #%ld as deleted\n", node->id);
+    printf("mark node #" ITEM_ID_FORMAT " as deleted\n", node->id);
     node->flags |= OSM_FLAG_DELETED;
   } else {
-    printf("permanently delete node #%ld\n", node->id);
+    printf("permanently delete node #" ITEM_ID_FORMAT "\n", node->id);
 
     /* remove it from the chain */
     node_t **cnode = &osm->node;
@@ -1615,16 +1689,16 @@ relation_chain_t *osm_node_to_relation(osm_t *osm, node_t *node) {
 
     member_t *member = relation->member;
     while(member) {
-      switch(member->type) {
+      switch(member->object.type) {
       case NODE:
 	/* nodes are checked directly */
-	if(member->node == node)
+	if(member->object.node == node)
 	  is_member = TRUE;
 	break;
 
       case WAY: {
 	/* ways have to be checked for the nodes they consist of */
-	node_chain_t *chain = member->way->node_chain;
+	node_chain_t *chain = member->object.way->node_chain;
 	while(chain && !is_member) {
 	  if(chain->node == node)
 	    is_member = TRUE;
@@ -1662,10 +1736,10 @@ relation_chain_t *osm_way_to_relation(osm_t *osm, way_t *way) {
 
     member_t *member = relation->member;
     while(member) {
-      switch(member->type) {
+      switch(member->object.type) {
       case WAY: {
 	/* ways can be check directly */
-	if(member->way == way)
+	if(member->object.way == way)
 	  is_member = TRUE;
       } break;
       
@@ -1727,15 +1801,15 @@ gboolean osm_position_within_bounds(osm_t *osm, gint x, gint y) {
 /* be deleted */
 void osm_node_remove_from_relation(osm_t *osm, node_t *node) {
   relation_t *relation = osm->relation;
-  printf("removing node #%ld from all relations:\n", node->id);
+  printf("removing node #" ITEM_ID_FORMAT " from all relations:\n", node->id);
 
   while(relation) {
     member_t **member = &relation->member;
     while(*member) {
-      if(((*member)->type == NODE) &&
-	 ((*member)->node == node)) {
+      if(((*member)->object.type == NODE) &&
+	 ((*member)->object.node == node)) {
 
-	printf("  from relation #%ld\n", relation->id);
+	printf("  from relation #" ITEM_ID_FORMAT "\n", relation->id);
 	
 	member_t *cur = *member;
 	*member = (*member)->next;
@@ -1752,15 +1826,15 @@ void osm_node_remove_from_relation(osm_t *osm, node_t *node) {
 /* remove the given way from all relations */
 void osm_way_remove_from_relation(osm_t *osm, way_t *way) {
   relation_t *relation = osm->relation;
-  printf("removing way #%ld from all relations:\n", way->id);
+  printf("removing way #" ITEM_ID_FORMAT " from all relations:\n", way->id);
 
   while(relation) {
     member_t **member = &relation->member;
     while(*member) {
-      if(((*member)->type == WAY) &&
-	 ((*member)->way == way)) {
+      if(((*member)->object.type == WAY) &&
+	 ((*member)->object.way == way)) {
 
-	printf("  from relation #%ld\n", relation->id);
+	printf("  from relation #" ITEM_ID_FORMAT "\n", relation->id);
 	
 	member_t *cur = *member;
 	*member = (*member)->next;
@@ -1778,14 +1852,17 @@ relation_t *osm_relation_new(void) {
   printf("Creating new relation\n");
 
   relation_t *relation = g_new0(relation_t, 1);
+  relation->version = 1;
   relation->visible = TRUE;
   relation->flags = OSM_FLAG_NEW;
   relation->time = time(NULL);
 
+#ifndef API06
   /* add created_by tag */
   relation->tag = g_new0(tag_t, 1);
   relation->tag->key = g_strdup("created_by");
   relation->tag->value = g_strdup(PACKAGE " v" VERSION);
+#endif
 
   return relation;
 }
@@ -1808,8 +1885,8 @@ void osm_way_delete(osm_t *osm, icon_t **icon,
 
   /* new ways aren't stored on the server and are just deleted permanently */
   if(way->flags & OSM_FLAG_NEW) {
-    printf("About to delete NEW way #%ld -> force permanent delete\n", 
-	   way->id);
+    printf("About to delete NEW way #" ITEM_ID_FORMAT 
+	   " -> force permanent delete\n", way->id);
     permanently = TRUE;
   }
 
@@ -1818,7 +1895,7 @@ void osm_way_delete(osm_t *osm, icon_t **icon,
   while(*chain) {
 
     (*chain)->node->ways--;
-    printf("checking node #%ld (still used by %d)\n", 
+    printf("checking node #" ITEM_ID_FORMAT " (still used by %d)\n", 
 	   (*chain)->node->id, (*chain)->node->ways);
 
     /* this node must only be part of this way */
@@ -1844,10 +1921,10 @@ void osm_way_delete(osm_t *osm, icon_t **icon,
   way->node_chain = NULL;
 
   if(!permanently) {
-    printf("mark way #%ld as deleted\n", way->id);
+    printf("mark way #" ITEM_ID_FORMAT " as deleted\n", way->id);
     way->flags |= OSM_FLAG_DELETED;
   } else {
-    printf("permanently delete way #%ld\n", way->id);
+    printf("permanently delete way #" ITEM_ID_FORMAT "\n", way->id);
 
     /* remove it from the chain */
     way_t **cway = &osm->way;
@@ -1872,8 +1949,8 @@ void osm_relation_delete(osm_t *osm, relation_t *relation,
   /* new relations aren't stored on the server and are just */
   /* deleted permanently */
   if(relation->flags & OSM_FLAG_NEW) {
-    printf("About to delete NEW relation #%ld -> force permanent delete\n", 
-	   relation->id);
+    printf("About to delete NEW relation #" ITEM_ID_FORMAT 
+	   " -> force permanent delete\n", relation->id);
     permanently = TRUE;
   }
 
@@ -1881,10 +1958,10 @@ void osm_relation_delete(osm_t *osm, relation_t *relation,
   /* don't have any reference to the relation they are part of */
 
   if(!permanently) {
-    printf("mark relation #%ld as deleted\n", relation->id);
+    printf("mark relation #" ITEM_ID_FORMAT " as deleted\n", relation->id);
     relation->flags |= OSM_FLAG_DELETED;
   } else {
-    printf("permanently delete relation #%ld\n", relation->id);
+    printf("permanently delete relation #" ITEM_ID_FORMAT "\n", relation->id);
 
     /* remove it from the chain */
     relation_t **crelation = &osm->relation;
@@ -1903,7 +1980,7 @@ void osm_relation_delete(osm_t *osm, relation_t *relation,
   }
 }
 
-void osm_way_revert(way_t *way) {
+void osm_way_reverse(way_t *way) {
   node_chain_t *new = NULL;
 
   /* walk old chain first to last */
@@ -1921,6 +1998,138 @@ void osm_way_revert(way_t *way) {
   way->node_chain = new;
 }
 
+static const char *DS_ONEWAY_FWD = "yes";
+static const char *DS_ONEWAY_REV = "-1";
+static const char *DS_LEFT_SUFFIX = ":left";
+static const char *DS_RIGHT_SUFFIX = ":right";
+
+/* Reverse direction-sensitive tags like "oneway". Marks the way as dirty if
+ * anything is changed, and returns the number of flipped tags. */
+
+guint
+osm_way_reverse_direction_sensitive_tags (way_t *way) {
+  tag_t *tag = way->tag;
+  guint n_tags_altered = 0;
+  while (tag != NULL) {
+    char *lc_key = g_ascii_strdown(tag->key, -1);
+    char *lc_value = g_ascii_strdown(tag->value, -1);
+    
+    if (strcmp(lc_key, "oneway") == 0) {
+      // oneway={yes/true/1/-1} is unusual.
+      // Favour "yes" and "-1".
+      if ((strcmp(lc_value, DS_ONEWAY_FWD) == 0) ||
+          (strcmp(lc_value, "true") == 0) ||
+          (strcmp(lc_value, "1") == 0)) {
+        g_free(tag->value);
+        tag->value = g_strdup(DS_ONEWAY_REV);
+        n_tags_altered++;
+      }
+      else if (strcmp(lc_value, DS_ONEWAY_REV) == 0) {
+        g_free(tag->value);
+        tag->value = g_strdup(DS_ONEWAY_FWD);
+        n_tags_altered++;
+      }
+      else {
+        printf("warning: unknown tag: %s=%s\n", tag->key, tag->value);
+      }
+    }
+
+    // :left and :right suffixes
+    else if (g_str_has_suffix(lc_key, DS_LEFT_SUFFIX)) {
+      char *key_old = tag->key;
+      char *lastcolon = rindex(key_old, ':');
+      g_assert(lastcolon != NULL);
+      *lastcolon = '\000';
+      tag->key = g_strconcat(key_old, DS_RIGHT_SUFFIX, NULL);
+      *lastcolon = ':';
+      g_free(key_old);
+      n_tags_altered++;
+    }
+    else if (g_str_has_suffix(lc_key, DS_RIGHT_SUFFIX)) {
+      char *key_old = tag->key;
+      char *lastcolon = rindex(key_old, ':');
+      g_assert(lastcolon != NULL);
+      *lastcolon = '\000';
+      tag->key = g_strconcat(key_old, DS_LEFT_SUFFIX, NULL);
+      *lastcolon = ':';
+      g_free(key_old);
+      n_tags_altered++;
+    }
+
+    g_free(lc_key);
+    g_free(lc_value);
+    tag = tag->next;
+  }
+  if (n_tags_altered > 0) {
+    way->flags |= OSM_FLAG_DIRTY;
+  }
+  return n_tags_altered;
+}
+
+/* Reverse a way's role within relations where the role is direction-sensitive.
+ * Returns the number of roles flipped, and marks any relations changed as
+ * dirty. */
+
+static const char *DS_ROUTE_FORWARD = "forward";
+static const char *DS_ROUTE_REVERSE = "reverse";
+
+guint
+osm_way_reverse_direction_sensitive_roles(osm_t *osm, way_t *way) {
+  relation_chain_t *rel_chain0, *rel_chain;
+  rel_chain0 = rel_chain = osm_way_to_relation(osm, way);
+  guint n_roles_flipped = 0;
+
+  for (; rel_chain != NULL; rel_chain = rel_chain->next) {
+    char *type = osm_tag_get_by_key(rel_chain->relation->tag, "type");
+    
+    // Route relations; http://wiki.openstreetmap.org/wiki/Relation:route
+    if (strcasecmp(type, "route") == 0) {
+      
+      // First find the member corresponding to our way:
+      member_t *member = rel_chain->relation->member;
+      for (; member != NULL; member = member->next) {
+        if (member->object.type == WAY) {
+          if (member->object.way == way)
+            break;
+        }
+        if (member->object.type == WAY_ID) {
+          if (member->object.id == way->id)
+            break;
+        }
+      }
+      g_assert(member);  // osm_way_to_relation() broken?
+      
+      // Then flip its role if it's one of the direction-sensitive ones
+      if (member->role == NULL) {
+        printf("null role in route relation -> ignore\n");
+      }
+      else if (strcasecmp(member->role, DS_ROUTE_FORWARD) == 0) {
+        g_free(member->role);
+        member->role = g_strdup(DS_ROUTE_REVERSE);
+        rel_chain->relation->flags |= OSM_FLAG_DIRTY;
+        ++n_roles_flipped;
+      }
+      else if (strcasecmp(member->role, DS_ROUTE_REVERSE) == 0) {
+        g_free(member->role);
+        member->role = g_strdup(DS_ROUTE_FORWARD);
+        rel_chain->relation->flags |= OSM_FLAG_DIRTY;
+        ++n_roles_flipped;
+      }
+
+      // TODO: what about numbered stops? Guess we ignore them; there's no
+      // consensus about whether they should be placed on the way or to one side
+      // of it.
+
+    }//if-route
+
+
+  }
+  if (rel_chain0) {
+    g_free(rel_chain0);
+  }
+  return n_roles_flipped;
+}
+
 node_t *osm_way_get_first_node(way_t *way) {
   node_chain_t *chain = way->node_chain;
   if(!chain) return NULL;
@@ -1980,7 +2189,7 @@ tag_t *osm_tags_copy(tag_t *src_tag, gboolean update_creator) {
 }
 
 /* return plain text of type */
-char *osm_type_string(type_t type) {
+char *osm_object_type_string(object_t *object) {
   const struct { type_t type; char *name; } types[] = {
     { ILLEGAL,     "illegal" },
     { NODE,        "node" },
@@ -1994,80 +2203,81 @@ char *osm_type_string(type_t type) {
 
   int i;
   for(i=0;types[i].name;i++) 
-    if(type == types[i].type)
+    if(object->type == types[i].type)
       return types[i].name;
 
   return NULL;
 }
 
-char *osm_object_string(type_t type, void *object) {
-  char *type_str = osm_type_string(type);
+char *osm_object_string(object_t *object) {
+  char *type_str = osm_object_type_string(object);
 
   if(!object) 
     return g_strdup_printf("%s #<invalid>", type_str);
 
-  switch(type) {
+  switch(object->type) {
   case ILLEGAL:
     return g_strdup_printf("%s #<unspec>", type_str);
     break;
   case NODE:
-    return g_strdup_printf("%s #%ld", type_str, ((node_t*)object)->id);
+    return g_strdup_printf("%s #" ITEM_ID_FORMAT, type_str, object->node->id);
     break;
   case WAY:
-    return g_strdup_printf("%s #%ld", type_str, ((way_t*)object)->id);
+    return g_strdup_printf("%s #" ITEM_ID_FORMAT, type_str, object->way->id);
     break;
   case RELATION:
-    return g_strdup_printf("%s #%ld", type_str, ((relation_t*)object)->id);
+    return g_strdup_printf("%s #" ITEM_ID_FORMAT, type_str, 
+			   object->relation->id);
     break;
   case NODE_ID:
   case WAY_ID:
   case RELATION_ID:
-    return g_strdup_printf("%s #%ld", type_str, ((item_id_t)object));
+    return g_strdup_printf("%s #" ITEM_ID_FORMAT, type_str, object->id);
     break;
   }
   return NULL;
 }
 
-char *osm_id_string(type_t type, void *object) {
+char *osm_object_id_string(object_t *object) {
   if(!object) return NULL;
 
-  switch(type) {
+  switch(object->type) {
   case ILLEGAL:
     return NULL;
     break;
   case NODE:
-    return g_strdup_printf("#%ld", ((node_t*)object)->id);
+    return g_strdup_printf("#"ITEM_ID_FORMAT, object->node->id);
     break;
   case WAY:
-    return g_strdup_printf("#%ld", ((way_t*)object)->id);
+    return g_strdup_printf("#"ITEM_ID_FORMAT, object->way->id);
     break;
   case RELATION:
-    return g_strdup_printf("#%ld", ((relation_t*)object)->id);
+    return g_strdup_printf("#"ITEM_ID_FORMAT, object->relation->id);
     break;
   case NODE_ID:
   case WAY_ID:
   case RELATION_ID:
-    return g_strdup_printf("#%ld", ((item_id_t)object));
+    return g_strdup_printf("#"ITEM_ID_FORMAT, object->id);
     break;
   }
   return NULL;
 }
 
-tag_t *osm_object_get_tags(type_t type, void *object) {
+tag_t *osm_object_get_tags(object_t *object) {
   if(!object) return NULL;
 
-  switch(type) {
+  switch(object->type) {
   case ILLEGAL:
     return NULL;
     break;
   case NODE:
-    return ((node_t*)object)->tag;
+    return object->node->tag;
     break;
   case WAY:
-    return ((way_t*)object)->tag;
+    return object->way->tag;
     break;
   case RELATION:
-    return ((relation_t*)object)->tag;
+    return object->relation->tag;
     break;
   case NODE_ID:
   case WAY_ID:
@@ -2089,4 +2299,28 @@ gint osm_relation_members_num(relation_t *relation) {
   return num;
 }
 
+void osm_object_set_flags(object_t *object, int set, int clr) {
+
+  switch(object->type) {
+  case NODE:
+    object->node->flags |=  set;
+    object->node->flags &= ~clr;
+    break;
+
+  case WAY:
+    object->way->flags |=  set;
+    object->way->flags &= ~clr;
+    break;
+
+  case RELATION:
+    object->relation->flags |=  set;
+    object->relation->flags &= ~clr;
+    break;
+
+  default:
+    g_assert(0);
+    break;
+  }
+}
+
 // vim:et:ts=8:sw=2:sts=2:ai
