@@ -27,7 +27,7 @@
 #define COLOR_ERR  "red"
 #define COLOR_OK   "darkgreen"
 
-#undef NO_EXPECT
+#define NO_EXPECT
 
 static struct http_message_s {
   int id;
